@@ -64,6 +64,14 @@ class FileStat(abc.ABC):
 
 
 class ForkedError(RuntimeError):
+    '''An error class when PFIO found the process forked.
+
+    If an FS object is not "lazy", any object usage detects process
+    fork and raises this ``ForkedError`` as soon as possible at the
+    child process. The parent process may or may not run well,
+    depending on the ``FS`` implementation.
+
+    '''
     pass
 
 
@@ -73,7 +81,6 @@ class FS(abc.ABC):
     '''
 
     _cwd = ''
-    pid = os.getpid()
 
     def __init__(self):
         self.pid = os.getpid()
@@ -125,6 +132,7 @@ class FS(abc.ABC):
 
     @property
     def is_forked(self):
+        assert hasattr(self, 'pid')
         return self.pid != os.getpid()
 
     def close(self) -> None:
@@ -264,7 +272,7 @@ class FS(abc.ABC):
 
 
 @contextlib.contextmanager
-def open_url(url: str, mode: str = 'r') -> 'IOBase':
+def open_url(url: str, mode: str = 'r', **kwargs) -> 'IOBase':
     '''Opens a file regardless of the backend FS type
 
     ``url`` must be compliant with URL standard in
@@ -275,14 +283,14 @@ def open_url(url: str, mode: str = 'r') -> 'IOBase':
            f.read()
 
     .. note:: Some FS resouces won't be closed when using this
-        functionality.
+        functionality. See ``from_url`` for keyword arguments.
 
     Returns:
         a FileObject that must be closed.
 
     '''
     dirname, filename = os.path.split(url)
-    with from_url(dirname) as fs:
+    with from_url(dirname, **kwargs) as fs:
         with fs.open(filename, mode) as fp:
             yield fp
 
@@ -290,22 +298,74 @@ def open_url(url: str, mode: str = 'r') -> 'IOBase':
 def from_url(url: str, **kwargs) -> 'FS':
     '''Factory pattern implementation, creates FS from URI
 
+    If ``force_type`` is set with archive type, not scheme,
+    it ignores the suffix and tries the specified archive
+    format by opening the blob file.
+
+    If ``force_type`` is set with scheme type, the FS will
+    built from it accordingly. The URL path is supposed to
+    be a directory for file systems or a path prefix for S3.
+
+    Arguments:
+        url (str): A URL string compliant with RFC 1738.
+
+        force_type (str): Force type of FS to be returned.
+            One of "zip", "hdfs", "s3", or "file", returned
+            respectively. Default is ``"file"``.
+
+        create (bool): Create the specified path doesn't exist.
+
     .. note:: Some FS resouces won't be closed when using this
         functionality.
 
     '''
     parsed = urlparse(url)
+    force_type = kwargs.pop('force_type', None)
 
     if parsed.scheme:
         scheme = parsed.scheme
     else:
         scheme = 'file'  # Default is local
 
-    if parsed.path.endswith('.zip'):
+    # When ``force_type`` is defined, it must be equal with given one.
+    if force_type is not None and force_type != "zip":
+        if force_type != scheme:
+            raise ValueError("URL scheme mismatch with forced type")
+
+    def _zip_check_create_not_supported():
+        if kwargs.get('create', False):
+            msg = '"create" option is not supported for Zip FS.'
+            raise ValueError(msg)
+
+    # force_type \ suffix | .zip    | other
+    # --------------------+---------+------
+    #                 zip | ok      | try zip
+    #             (other) | try dir | try dir
+    #                None | try zip | try dir
+    if force_type == 'zip':
+        _zip_check_create_not_supported()
         dirname, filename = os.path.split(parsed.path)
+        fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)
+        fs = fs.open_zip(filename)
+
+    elif force_type is None:
+        if parsed.path.endswith('.zip'):
+            _zip_check_create_not_supported()
+            dirname, filename = os.path.split(parsed.path)
+            fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)
+            fs = fs.open_zip(filename)
+        else:
+            dirname = parsed.path
+            fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)
+
     else:
         dirname = parsed.path
+        fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)
 
+    return fs
+
+
+def _from_scheme(scheme, dirname, kwargs, bucket=None):
     if scheme == 'file':
         from .local import Local
         fs = Local(dirname, **kwargs)
@@ -314,15 +374,10 @@ def from_url(url: str, **kwargs) -> 'FS':
         fs = Hdfs(dirname, **kwargs)
     elif scheme == 's3':
         from .s3 import S3
-
-        fs = S3(bucket=parsed.netloc, prefix=dirname, **kwargs)
-
+        fs = S3(bucket=bucket, prefix=dirname, **kwargs)
     else:
         raise RuntimeError("Scheme {} is not supported", scheme)
 
-    if parsed.path.endswith('.zip'):
-        return fs.open_zip(filename)
-
     return fs
 
 
