@@ -150,7 +150,7 @@ def _create_fs():
     configs = {}
     for e in root.findall('./property'):
         name = None
-        for c in e.getchildren():
+        for c in e:
             if c.tag == 'name':
                 name = c.text
             elif c.tag == 'value':
@@ -164,7 +164,7 @@ def _create_fs():
         # amont multiple name services from single HADOOP_CONF_DIR
         # conf. Thus we ignore fs.defaultFS and just take the very
         # first name service that appeared in hdfs-site.xml.
-        return HadoopFileSystem(nameservice)
+        return HadoopFileSystem(nameservice, 0)
 
     else:
         RuntimeError("No nameservice found.")
@@ -175,12 +175,21 @@ class Hdfs(FS):
 
     To use HDFS, PFIO requires ``$HADOOP_HOME`` predefined before
     initialization. If it is not defined, ``ARROW_LIBHDFS_DIR`` must
-    be defined instead. If ``$CLASSPATH`` will be needed in case
-    ``hdfs`` command is not available from ``$PATH``.
+    be defined instead. ``$CLASSPATH`` will be needed in case ``hdfs``
+    command is not available from ``$PATH``.
 
+    .. note:: With environment variable
+          ``KRB5_KTNAME=path/to/your.keytab`` set, ``hdfs``
+          handler automatically starts automatic and periodical
+          updating Kerberos ticket using `krbticket
+          <https://pypi.org/project/krbticket/>`_ . The update
+          frequency is every 10 minutes by default.
+    .. note::
+          Only the username in the first entry in The
+          keytab will be used to update the Kerberos ticket.
     '''
 
-    def __init__(self, cwd=None):
+    def __init__(self, cwd=None, create=False, **_):
         super().__init__()
         self._fs = _create_fs()
         assert self._fs is not None
@@ -194,6 +203,12 @@ class Hdfs(FS):
             else:
                 self.cwd = os.path.join(self.cwd, cwd)
 
+        if not self.isdir(''):
+            if create:
+                self.makedirs('')
+            else:
+                raise ValueError('{} must be a directory'.format(self.cwd))
+
     def _get_principal_name(self):
         # get the default principal name from `klist` cache
         principal_name = _get_principal_name_from_klist()
@@ -272,7 +287,7 @@ class Hdfs(FS):
         self._checkfork()
         path = os.path.join(self.cwd, path)
         info = self._fs.get_file_info(path)
-        return not info.is_file
+        return info.type == FileType.Directory
 
     def mkdir(self, path: str, *args, dir_fd=None):
         self._checkfork()
