@@ -43,20 +43,27 @@ class LocalFileStat(FileStat):
 
 
 class Local(FS):
-    def __init__(self, cwd=None):
+    def __init__(self, cwd=None, create=False, **_):
         super().__init__()
+
         if cwd is None:
             self._cwd = ''
         else:
             self._cwd = cwd
 
+        if not self.isdir(''):
+            if create:
+                os.makedirs(self._cwd)
+            else:
+                raise ValueError('{} must be a directory'.format(self._cwd))
+
     @property
     def cwd(self):
-        if self._cwd is None:
-            return os.getcwd()
-        else:
+        if self._cwd:
             return self._cwd
 
+        return os.getcwd()
+
     def open(self, file_path, mode='r',
              buffering=-1, encoding=None, errors=None,
              newline=None, closefd=True, opener=None):
