@@ -9,6 +9,16 @@ from botocore.exceptions import ClientError
 
 from .fs import FS, FileStat
 
+DEFAULT_MAX_BUFFER_SIZE = 16 * 1024 * 1024
+
+
+def _normalize_key(key: str) -> str:
+    key = os.path.normpath(key)
+    if key.startswith("/"):
+        return key[1:]
+    else:
+        return key
+
 
 class S3ObjectStat(FileStat):
     def __init__(self, key, head):
@@ -22,8 +32,20 @@ class S3ObjectStat(FileStat):
         return False
 
 
-class _ObjectReader:
+class S3PrefixStat(FileStat):
+    def __init__(self, key):
+        self.filename = key
+        self.last_modified = 0
+        self.size = -1
+
+    def isdir(self):
+        return True
+
+
+class _ObjectReader(io.RawIOBase):
     def __init__(self, client, bucket, key, mode, kwargs):
+        super(_ObjectReader, self).__init__()
+
         self.client = client
         self.bucket = bucket
         self.key = key
@@ -65,6 +87,9 @@ class _ObjectReader:
 
         return data
 
+    def readline(self):
+        raise NotImplementedError()
+
     def close(self):
         self._closed = True
 
@@ -92,23 +117,42 @@ class _ObjectReader:
     def seekable(self):
         return True
 
+    def tell(self):
+        return self.pos
+
+    def truncate(self, size=None):
+        raise io.UnsupportedOperation('truncate')
+
     def seek(self, pos, whence=io.SEEK_SET):
         if whence in [0, io.SEEK_SET]:
-            self.pos = pos
+            if pos < 0:
+                raise OSError(22, "[Errno 22] Invalid argument")
         elif whence in [1, io.SEEK_CUR]:
-            self.pos += pos
+            pos += self.pos
         elif whence in [2, io.SEEK_END]:
-            self.pos += pos
+            pos += self.content_length
+        else:
+            raise ValueError('Wrong whence value: {}'.format(whence))
 
-        if self.content_length < self.pos:
-            self.pos = self.content_length
-        if self.pos < 0:
-            raise OSError()
+        if pos < 0:
+            raise OSError(22, "[Errno 22] Invalid argument")
+        self.pos = pos
         return self.pos
 
     def writable(self):
         return False
 
+    def write(self, data):
+        raise io.UnsupportedOperation('not writable')
+
+    def readall(self):
+        return self.read(-1)
+
+    def readinto(self, b):
+        buf = self.read(len(b))
+        b[:len(buf)] = buf
+        return len(buf)
+
 
 class _ObjectWriter:
     def __init__(self, client, bucket, key, mode, mpu_chunksize, kwargs):
@@ -188,13 +232,16 @@ class _ObjectWriter:
             self._flush()
             # DO: MPU
             c = self.client
-            max_parts = len(self.parts)
+            max_parts = len(self.parts) + 1
             res = c.list_parts(Bucket=self.bucket,
                                Key=self.key,
                                UploadId=self.mpu_id, MaxParts=max_parts)
 
             if res['IsTruncated']:
-                raise RuntimeError('truncated.')
+                next_part = res['NextPartNumberMarker']
+                raise RuntimeError('Unexpectedly truncated: ' +
+                                   'next={}/maxparts={}'.format(next_part,
+                                                                max_parts))
 
             parts = [{'ETag': part['ETag'], 'PartNumber': part['PartNumber']}
                      for part in res.get('Parts', [])]
@@ -245,20 +292,34 @@ class S3(FS):
     - ``aws_secret_access_key``, ``AWS_SECRET_ACCESS_KEY``
     - ``endpoint``, ``S3_ENDPOINT``
 
+    It supports buffering when opening a file in binary read mode ("rb").
+    When ``buffering`` is set to -1 (default), the buffer size will be
+    the size of the file or ``pfio.v2.S3.DEFAULT_MAX_BUFFER_SIZE``,
+    whichever smaller.
+    ``buffering=0`` disables buffering, and ``buffering>0`` forcibly sets the
+    specified value as the buffer size in bytes.
     '''
 
     def __init__(self, bucket, prefix=None,
                  endpoint=None, create_bucket=False,
                  aws_access_key_id=None,
                  aws_secret_access_key=None,
-                 mpu_chunksize=32*1024*1024):
+                 mpu_chunksize=32*1024*1024,
+                 buffering=-1,
+                 create=False,
+                 **_):
+        super().__init__()
         self.bucket = bucket
         if prefix is not None:
             self.cwd = prefix
         else:
             self.cwd = ''
 
+        # In S3, create flag can be disregarded
+        del create
+
         self.mpu_chunksize = mpu_chunksize
+        self.buffering = buffering
 
         # boto3.set_stream_logger()
 
@@ -316,10 +377,15 @@ class S3(FS):
             raise io.UnsupportedOperation('Read-write mode is not supported')
 
         path = os.path.join(self.cwd, path)
+        path = _normalize_key(path)
         if 'r' in mode:
             obj = _ObjectReader(self.client, self.bucket, path, mode, kwargs)
             if 'b' in mode:
-                obj = io.BufferedReader(obj)
+                if self.buffering:
+                    bs = self.buffering
+                    if bs < 0:
+                        bs = min(obj.content_length, DEFAULT_MAX_BUFFER_SIZE)
+                    obj = io.BufferedReader(obj, buffer_size=bs)
             else:
                 obj = io.TextIOWrapper(obj)
 
@@ -342,10 +408,11 @@ class S3(FS):
 
         '''
         self._checkfork()
-        key = os.path.normpath(os.path.join(self.cwd, prefix))
+        key = os.path.join(self.cwd, prefix)
+        key = _normalize_key(key)
         if key == '.':
             key = ''
-        if key:
+        elif key != '' and not key.endswith('/'):
             key += '/'
         if '/../' in key or key.startswith('..'):
             raise ValueError('Invalid S3 key: {} as {}'.format(prefix, key))
@@ -363,7 +430,7 @@ class S3(FS):
         for res in iterator:
             # print(res)
             for common_prefix in res.get('CommonPrefixes', []):
-                yield common_prefix['Prefix']
+                yield common_prefix['Prefix'][len(key):]
             for content in res.get('Contents', []):
                 yield content['Key'][len(key):]
 
@@ -373,32 +440,61 @@ class S3(FS):
         '''
         self._checkfork()
         key = os.path.join(self.cwd, path)
-        res = self.client.head_object(Bucket=self.bucket,
-                                      Key=key)
-        if res.get('DeleteMarker'):
-            raise FileNotFoundError()
+        key = _normalize_key(key)
+        try:
+            res = self.client.head_object(Bucket=self.bucket,
+                                          Key=key)
+            if res.get('DeleteMarker'):
+                raise FileNotFoundError()
 
-        return S3ObjectStat(key, res)
+            return S3ObjectStat(key, res)
+        except ClientError as e:
+            if e.response['Error']['Code'] == '404':
+                if self.isdir(path):
+                    return S3PrefixStat(key)
+                raise FileNotFoundError()
+            else:
+                raise e
 
     def isdir(self, file_path: str):
-        '''Does nothing
-
-        .. note:: AWS S3 does not have concept of directory tree; what
-           this function (and ``mkdir()`` and ``makedirs()`` should do
-           and return? To be strict, it would be straightforward to
-           raise ``io.UnsupportedOperation`` exception. But it just
-           breaks users' applications that except quasi-compatible
-           behaviour. Thus, imitating other file systems, like
-           returning boolean or ``None`` would be nicer.
+        '''Imitate isdir by handling common prefix ending with "/" as directory
 
+        AWS S3 does not have concept of directory tree, but this class
+        imitates other file systems to increase compatibility.
         '''
-        # raise io.UnsupportedOperation("S3 doesn't have directory")
-        pass
+        self._checkfork()
+        key = _normalize_key(os.path.join(self.cwd, file_path))
+        if key == '.':
+            key = ''
+        elif key.endswith('/'):
+            key = key[:-1]
+        if '/../' in key or key.startswith('..'):
+            raise ValueError('Invalid S3 key: {} as {}'.format(file_path, key))
+
+        if len(key) == 0:
+            return True
+
+        res = self.client.list_objects_v2(
+            Bucket=self.bucket,
+            Prefix=key,
+            Delimiter="/",
+            MaxKeys=1,
+        )
+        for common_prefix in res.get('CommonPrefixes', []):
+            if common_prefix['Prefix'] == key + "/":
+                return True
+        return False
 
     def mkdir(self, file_path: str, mode=0o777, *args, dir_fd=None):
         '''Does nothing
 
-        .. note:: see discussion in ``isdir()``.
+        .. note:: AWS S3 does not have concept of directory tree; what
+           this function (and ``makedirs()``) should do
+           and return? To be strict, it would be straightforward to
+           raise ``io.UnsupportedOperation`` exception. But it just
+           breaks users' applications that except quasi-compatible
+           behaviour. Thus, imitating other file systems, like
+           returning ``None`` would be nicer.
         '''
         # raise io.UnsupportedOperation("S3 doesn't have directory")
         pass
@@ -406,7 +502,7 @@ class S3(FS):
     def makedirs(self, file_path: str, mode=0o777, exist_ok=False):
         '''Does nothing
 
-        .. note:: see discussion in ``isdir()``.
+        .. note:: see discussion in ``mkdir()``.
         '''
         # raise io.UnsupportedOperation("S3 doesn't have directory")
         pass
@@ -419,11 +515,14 @@ class S3(FS):
         self._checkfork()
         try:
             key = os.path.join(self.cwd, file_path)
+            key = _normalize_key(key)
             res = self.client.head_object(Bucket=self.bucket,
                                           Key=key)
             return not res.get('DeleteMarker')
         except ClientError as e:
             if e.response['Error']['Code'] == '404':
+                if self.isdir(file_path):
+                    return True
                 return False
             else:
                 raise e
@@ -436,24 +535,34 @@ class S3(FS):
 
         '''
         self._checkfork()
-        source = {'Bucket': self.bucket, 'Key': os.path.join(self.cwd, src)}
+        source = {
+            'Bucket': self.bucket,
+            'Key': _normalize_key(os.path.join(self.cwd, src)),
+        }
         dst = os.path.join(self.cwd, dst)
+        dst = _normalize_key(dst)
         res = self.client.copy_object(Bucket=self.bucket,
                                       CopySource=source,
                                       Key=dst)
         if not res.get('CopyObjectResult'):
             # copy failed
             return
-        return self.remove(source.get('Key'))
+        return self.remove(src)
 
     def remove(self, file_path: str, recursive=False):
         '''Removes an object
 
+        It raises a FileNotFoundError when the specified file doesn't exist.
         '''
         if recursive:
             raise io.UnsupportedOperation("Recursive delete not supported")
 
+        if not self.exists(file_path):
+            msg = "No such S3 object: '{}'".format(file_path)
+            raise FileNotFoundError(msg)
+
         self._checkfork()
         key = os.path.join(self.cwd, file_path)
+        key = _normalize_key(key)
         return self.client.delete_object(Bucket=self.bucket,
                                          Key=key)
